Program COCOPAN (Input,Output,Links,Coord,DataFile,Statis,Graphics);

CONST   maxloc=200;      (* Maximum number of objects (localities) in a run.  *)
        maxlocPlus1=251; (* Equals maxloc + 1                                 *)
        maxgroupes=14;   (* Maximum number of groups + 1. If it becomes       *)
                         (* necessary to change this parameter, the alphabet  *)
                         (* in PROCEDURE "Init" should also be completed to   *)
                         (* provide "Maxgroupes" printing char.               *)
        maxvar=12;       (* Maximum number of alleles (variables) in a run    *)
        epsilon=10e-7;   (* A very small positive real number.                *)
        maxreel=1.70e38; (* Largest REAL number on a VAX mainframe computer.
                            It may have to be changed on other machines.      *)
        rad=57.295779513;(* Number of degrees in one radian.                  *)
        facteur=1.25;    (* Printing factor: horizontal char./vertical char.  *)
        maxlargeur=120;  (* Width of the printed map on the line printer.     *)
        maxlargp1=121;   (*                 Maxlargeur + 1                    *)
        nbbits=256;      (* Maximum length of a Pascal set on this computer.  *)
        nbbitsl1=255;    (*                 Nbbits -1                         *)
        maxmots=0;       (* Largest INTEGER in the ratio   maxloc / Nbbits    *)
        maxdegree=20;    (* Random tree algorithm: maximum number of neighbors  
                            allowed for any given data point, in file LINKS.  *)
        lang=1;          (* Parametre de langue:  1=Francais
                            Language parameter :  2=English                   *)
(* =====================================================================
COCOPAN stands for Contiguity-Constrained Permutational ANOVA.
This is version 1.1 of COCOPAN (December 1990). 
This program, with the Ring algorithm only, was formerly called PRECEL.

     This program carries out a one-way analysis of variance for 
spatially autocorrelated quantitative data when the classification 
criterion is a subdivision of the study area into nonoverlapping regions
(ex. countries, counties, language areas, geomorphological subdivisions 
of the study area, and so on). Several variables can be analyzed in a 
single run. The statistic subjected to permutation testing is the 
within-group sum of squares, so that the critical area is the lower tail
of the probability distribution. Probabilities are computed following 
Hope (1968).

Reference to the method:

Legendre, P., N.L. Oden, R.R. Sokal, A. Vaudor and J. Kim. 1990.
   Approximate analysis of variance of spatially autocorrelated
   regional data. J. Class. 7: 53-75.

Files:                                                                
   INPUT and OUTPUT are connected to the user's terminal.        
   LINKS is a list of pairs of neighbors. For instance, the file of link
      edges produced by programs AUTOCOR or CONNEXIONS can be used.                                                       
   DATAFILE is the main data file. The lines correspond to the  
      localities. The first N columns are the N variables to be 
      analyzed, while the last column contains the classification 
      criterion variable, coded from 1 to some appropriate INTEGER
      value smaller than Maxgroupes (see CONST above).              
   COORD is the list of geographic coordinates (X and Y) of the
      localities. It is needed only if one chooses to print the maps.
   GRAPHICS is the output file on which the maps are printed.
   STATIS is the output file containing detailed statistics.
   
A user's information file is available.
===================================================================== *)

TYPE    Ensemble        = set of 0..nbbitsl1;
        GrBit           = set of 0..maxgroupes;
        VecteurBit      = ARRAY[0..maxmots]of Ensemble;
        neighbours      = ARRAY[1..maxloc] of VecteurBit;
        matbit          = ARRAY[1..maxgroupes] of VecteurBit;
        intgr           = ARRAY[1..maxgroupes] of INTEGER;
        intobj          = ARRAY[0..maxlocPlus1] of INTEGER;
        datum           = ARRAY[1..maxloc] of REAL;
        data            = ARRAY[1..maxvar] of datum;
        pourvar         = ARRAY[1..maxvar] of REAL;
        pourgr          = ARRAY[1..maxgroupes] of REAL;
        stattype        = ARRAY[1..maxvar] of INTEGER;
        coordonnees     = ARRAY[1..2,1..maxloc] of REAL;
        resulttype      = ARRAY[1..maxvar] of intgr;
        reftype         = ARRAY[1..maxvar] of pourgr;
        rec             = RECORD re1,re2:REAL;end;
        NodePtr         = @NodeMap;
        NodeMap         = RECORD
                                Name        : INTEGER;
                                Used        : BOOLEAN;
                                Degree        : INTEGER;
                                Link        : ARRAY[1..maxdegree] OF NodePtr;
                                END;
        MAPType         = ARRAY[1..maxloc] OF NodePtr;
        FreqType        = ARRAY[1..MaxGroupes,1..maxloc] of INTEGER;
VAR     i,j,nobj,nbvar,nbgroupes,nbreussis,largeur,largeurp1,totalpasses,
        nbdept,nbforce,NbForces,nbforcereussis,nbmots, NbCartes, maxpasses, jj,
                MaxAbort,NofLink:INTEGER;
        statisset,GrapheSet:set of 1..6;
        action,Dessin,coccinelle,statisbool,VaudorMethod:boolean;
        x:REAL;
        dseed:REAL;
        voisins,faits,courant:neighbours;
        groupes:matbit;
        dejafait,vectinit:VecteurBit;
        copy1,copy2,seed:@VecteurBit;
        cardinalite,pathlength:intgr;
        cardgr,seedcopy,priorite,OriginalPl,PPPl,EQPl,PPDiam,EQDiam:@intgr;
        donnee:data;
        sumssq,sommessq:pourvar;
        pp,eq:stattype;
        ppm,eqm:resulttype;
        Freq:@FreqType;
        refr,pgdist:reftype;
        sumx2,sumdat:pourgr;
        pcoeff,OriginalDiam:@PourGr;
        position:coordonnees;
        coordx:ARRAY[1..2] of intobj;
        reference:intobj;
        deplacement:@Neighbours;
        alfabet:ARRAY[0..maxgroupes] of char;
        Links,Graphics,DataFile,coord,statis:TEXT;
        OriginalGroup              : @IntObj;
        MAP                        : @MAPType;


        PROCEDURE Erreur(i:integer);
        begin
        if lang=1 then Case i of
                1:writeln(' Fichier de donnees vide.');
                2:writeln(' Vous devez fournir un fichier de liens.');
                3:writeln(' Vous devez fournir un fichier de coordonnees.');
                4:writeln(' Fichier de donnees incomplet.');
                5:writeln(' Fichier de coordonnees incomplet.');
                6:writeln(' Pas plus de ',Maxdegree:2,' liens par localite ne',
                          ' sont autorises.');
                end
        else Case i of
                1:writeln(' Empty data file.');
                2:writeln(' File of links required.');
                3:writeln(' File of coordinates required.');
                4:writeln(' Incomplete data file.');
                5:writeln(' Incomplete file of coordinates.');
                6:writeln(' No more than ',Maxdegree:2,' edges per data point',
                          ' are permitted.');
                end;
        halt;
        end;

        FUNCTION GGUBFS (VAR DSEED:REAL):REAL;
        CONST    D2P31M=2147483647.0;
                 D2P31 =2147483648.0;
        VAR        A:REAL;
        BEGIN
        A:=16807*DSEED;
        DSEED:=A-TRUNC(A/D2P31M)*D2P31M;
        GGUBFS:=DSEED/D2P31;
        END;                    (* GGUBFS *)
         
        FUNCTION validebool:boolean;
        begin
        while (input@ <> '0')and(input@ <> 'O')and(input@ <>'Y')and
                (input@ <> 'N')and (input@ <> 'o')and(input@ <>'y')and
                (input@ <>'n') do get(input);
        validebool:=(input@ = '0')or(input@ = 'O')or(input@ = 'Y')or
                (input@ = 'o')or(input@ = 'y');
        readln;
        end;
         
        FUNCTION card(x:Ensemble):INTEGER;
        var c,i:INTEGER;
        begin
        c:=0;
        if x <> [] then for i:=0 to nbbitsl1 do if i in x then c:=c+1;
        card:=c;
        end;          (* card *)
         
        PROCEDURE identification(var fichier:text);
        begin
        writeln(fichier);
        if lang=1 then
          begin
          writeln(fichier,
          ' P r o g r a m m e   C O C O P A N  --  C a r t e s');
          writeln(fichier);
          writeln(fichier,
          ' (ANOVA par permutations sous contrainte de contiguite)');
          end 
        else begin
          writeln(fichier,' P r o g r a m   C O C O P A N  --  M a p');
          writeln(fichier);
          writeln(fichier,' (Contiguity-constrained permutational ANOVA)');
          end;
        writeln(fichier);
        writeln(fichier,' Reference:');
        writeln (fichier,
        ' Legendre, P., N.L. Oden, R.R. Sokal, A. Vaudor and J. Kim. 1990.');
        writeln(fichier,
        '    Approximate analysis of variance of spatially autocorrelated');
        writeln(fichier,
        '    regional data. J. Class. 7: 53-75.');
        writeln(fichier);
        if lang=1 then
        begin
        writeln(fichier,' Credits -- ');
        writeln(fichier,
        ' Programme et algorithme des cercles concentriques: Alain Vaudor,');
        writeln(fichier,' Departement de sciences biologiques,');
        writeln(fichier,' Universite de Montreal,');
        writeln(fichier,' C.P. 6128, Succursale A,');
        writeln(fichier,' Montreal, Quebec H3C 3J7.');
        writeln(fichier);
        writeln(fichier,
        ' Algorithme de l''arborescence aleatoire: Junhyong Kim,');
        writeln(fichier,' State University of New York at Stony Brook.');
        end else begin
        writeln(fichier,' Authorship -- ');
        writeln(fichier,' Program and Ring algorithm: Alain Vaudor,');
        writeln(fichier,' Departement de sciences biologiques,');
        writeln(fichier,' Universite de Montreal,');
        writeln(fichier,' C.P. 6128, Succursale A,');
        writeln(fichier,' Montreal, Quebec H3C 3J7.');
        writeln(fichier);
        writeln(fichier,' Random Tree algorithm: Junhyong Kim,');
        writeln(fichier,' State University of New York at Stony Brook.');
        end;
        writeln(fichier);
        end;
         
        FUNCTION vide(var m:VecteurBit):boolean;
        var i:INTEGER;
            b:boolean;
        begin
        b:=true;
        for i:=0 to nbmots do if m[i]<>[] then b:=false;
        vide:=b;
        end;
         
        FUNCTION membre(ind:INTEGER;var mm:VecteurBit):boolean;
        var i,j:INTEGER;
        begin
        i:=ind div nbbits;
        j:=ind mod nbbits;
        membre:=j in mm[i];
        end;
         
        PROCEDURE nullvec(var m:VecteurBit);
        var i:INTEGER;
        begin
        for i:=0 to nbmots do m[i]:=[];
        end;
         
        PROCEDURE union(var m1,m2:VecteurBit);
        var i:INTEGER;
        begin
        for i:=0 to nbmots do m1[i]:=m1[i]+m2[i];
        end;
         
        PROCEDURE intersection(var m0,m1,m2:VecteurBit);
        var i:INTEGER;
        begin
        for i:=0 to nbmots do m0[i]:=m1[i]*m2[i];
        end;
         
        FUNCTION inclusegal(var m1,m2:VecteurBit):boolean;
        var i:INTEGER;
            b:boolean;
        begin
        b:=true;
        for i:=0 to nbmots do if not(m1[i] <= m2[i]) then b:=false;
        inclusegal:=b;
        end;
         
        FUNCTION egal(var m1,m2:VecteurBit):boolean;
        var i:INTEGER;
            b:boolean;
        begin
        b:=true;
        for i:=0 to nbmots do if m1[i] <> m2[i] then b:=false;
        egal:=b;
        end;
         
        PROCEDURE ajoute(ind:INTEGER;var m:VecteurBit);
        var i:INTEGER;
        begin
        i:=ind div nbbits;
        m[i]:=m[i]+[ind mod nbbits];
        end;
         
        PROCEDURE enleve(ind:INTEGER;var m:VecteurBit);
        var i:INTEGER;
        begin
        i:=ind div nbbits;
        m[i]:=m[i]-[ind mod nbbits];
        end;
         
        PROCEDURE difference(var m1,m2:VecteurBit);
        var i:INTEGER;
        begin
        for i:=0 to nbmots do m1[i]:=m1[i]-m2[i];
        end;
         
        PROCEDURE init;
        begin
        alfabet[ 0]:=' ';
        alfabet[ 1]:='A';
        alfabet[ 2]:='B';
        alfabet[ 3]:='C';
        alfabet[ 4]:='D';
        alfabet[ 5]:='E';
        alfabet[ 6]:='F';
        alfabet[ 7]:='G';
        alfabet[ 8]:='H';
        alfabet[ 9]:='I';
        alfabet[10]:='J';
        alfabet[11]:='K';
        alfabet[12]:='L';
        alfabet[13]:='M';
        alfabet[14]:='N';
        (* alfabet[15]:='o';
           alfabet[16]:='p';
           alfabet[17]:='r';
           alfabet[18]:='s';
           alfabet[19]:='t';
           alfabet[20]:='u'; *)
        end;
         
        FUNCTION cardvect(var mat:VecteurBit):INTEGER;
        var i,compte:INTEGER;
        begin
        compte:=0;
        for i:=0 to nbmots do compte:=compte+card(mat[i]);
        cardvect:=compte;
        end;
         
        PROCEDURE creefrontiere(var frontiere,carte:VecteurBit);
        var i:INTEGER;
        begin
        nullvec(frontiere);
        for i:=1 to nobj do if membre(i,carte) then union(frontiere,voisins[i]);
        difference(frontiere,carte);
        end;
         
        PROCEDURE premier(var i:INTEGER;var t:VecteurBit);
        begin
        i:=0;
        repeat i:=i+1; until membre(i,t);
        end;
         
        PROCEDURE ecritvect(var mat:VecteurBit;addr:INTEGER);
        var i:INTEGER;
        begin
        write(Graphics,' ');
        for i:=addr to addr+79 do
                begin
                if i <= nobj then
                   if membre(i,mat) then write(Graphics,'+')
                   else write(Graphics,'-');
                end;
        if addr+79 >=nobj then write(Graphics,cardvect(mat):5);
        writeln(Graphics);
        end;
         
        FUNCTION adresse(ind:INTEGER):INTEGER;
        var i:INTEGER;
        begin
        i:=0;
        repeat
                i:=i+1;
        until (membre(ind,groupes[i])) or(i=nbgroupes);
        if membre(ind,groupes[nbgroupes]) then adresse:=nbgroupes
        else adresse:=i mod nbgroupes;
        end;
         
        PROCEDURE carte;
        var i,j:INTEGER;
            mat:ARRAY[0..maxlargp1] of char;
         
                PROCEDURE tiret;
                var i:INTEGER;
                begin
                write(Graphics,' ');
                for i:=0 to largeurp1 do write(Graphics,'-');
                writeln(Graphics);
                end;
         
        begin
        if Dessin then
                begin
                page(Graphics);
                largeurp1:=largeur+1;
                mat[0]:='!';
                mat[largeurp1]:=mat[0];
                writeln(Graphics);
                writeln(Graphics);
                if lang=1 then writeln(Graphics,' Carte no ',nbreussis:3)
                else writeln(Graphics,' Map no. ',nbreussis:3);
                if 2 in GrapheSet then
                        begin
                        for i:=0 to nobj div 80 do
                                begin
                                for j:=1 to nbgroupes do 
                                ecritvect(groupes[j],i*80+1);
                                end;
                        writeln(Graphics);
                        writeln(Graphics);
                        end;
                if 3 in GrapheSet then
                        begin
                        for i:=1 to nobj do
                                begin
                                j:=Adresse(i);
                                if (j < 0) OR (j > NbGroupes) then j:=0;
                                write(Graphics,j:4);
                                if i mod 20 = 0 then writeln(Graphics);
                                end;
                        if Nobj Mod 20 <> 0 then writeln(Graphics);
                        writeln(Graphics);
                        end;
                if 1 in GrapheSet then
                        begin
                        tiret;
                        i:=1;
                        repeat
                                for j:=1 to largeur do mat[j]:=' ';
                                repeat
                                        mat[coordx[1][i]]:=
                                        alfabet[adresse(reference[i])];
                                        i:=i+1;
                                until (coordx[2][i-1] <> coordx[2][i]);
                                write(Graphics,' ');
                                for j:=largeurp1 downto 0 do 
                                        write(Graphics,mat[j]);
                                writeln(Graphics);
                                if i <= nobj then for j:=
                                   coordx[2][i-1]-1 downto coordx[2][i]+1 do
                                        begin
                                        write(Graphics,' !');
                                        write(Graphics,' ':largeur);
                                        writeln(Graphics,'!');
                                        end;
                        until i > nobj;
                        tiret;
                        end;
                end;
        end;  (* carte *)
         
        FUNCTION connexe:boolean;
        var i,j:INTEGER;
            fait,carte,frontiere:VecteurBit;
            connecte:boolean;
         
                PROCEDURE connexion(var mat:VecteurBit);
                var i,j:INTEGER;
                begin
                if vide(mat) then connecte:=false
                else
                        begin
                        i:=0;
                        premier(i,mat);
                        nullvec(fait);
                        ajoute(i,fait);
                        carte:=fait;
                        repeat
                                creefrontiere(frontiere,carte);
                                carte:=frontiere;
                                difference(carte,fait);
                                intersection(carte,carte,mat);
                                union(fait,carte);
                                if vide(carte) then connecte:=false;
                        until (egal(mat,fait)) or (not(connecte));
                        end;
                end;
         
        begin
        connecte:=true;
        for i:=1 to nbgroupes do if connecte then
                if cardinalite[i] > 1 then connexion(groupes[i]);
        connexe:=connecte;
        end;
         
        FUNCTION ssq(var donnee:datum):REAL;
        var i,j:INTEGER;
            sum,x,sommex,sommex2:REAL;
        begin
        sum:=0;
        for i:=1 to nbgroupes do
                begin
                sommex:=0;sommex2:=0;
                for j:=1 to nobj do if membre(j,groupes[i]) then
                        begin
                        x:=donnee[j];
                        sommex:=sommex+x;
                        sommex2:=sommex2+sqr(x);
                        end;
                x:=sommex2-sqr(sommex)/cardinalite[i];
                sumdat[i]:=sommex;
                sumx2[i]:=x;
                sum:=sum+x;
                end;
        ssq:=sum;
        end;

        PROCEDURE compare(x,refr:REAL;var pp,eq:INTEGER);
        begin
        if x < refr - epsilon then pp:=pp+1
        else if x < refr + epsilon then eq:=eq+1;
        end;
                 
        PROCEDURE CalculeStatistiques;
        var i,j:INTEGER;
            x,y:REAL;
        begin
        for i:=1 to nbvar do
                begin
                x:=ssq(donnee[i]);
                sommessq[i]:=x;
                y:=sumssq[i];
                compare(x,y,pp[i],eq[i]);
                for j:=1 to nbgroupes do
                        compare(sumx2[j],refr[i][j],ppm[i][j],eqm[i][j]);
                pgdist[i]:=sumx2;
                end;
        end;
         
        PROCEDURE longueur;
        var i,j,k:INTEGER;
            tousfinis:boolean;
            fini:ARRAY[1..maxgroupes] of boolean;
            ones,temp1,temp2:VecteurBit;
         
        begin
        nullvec(ones);
        for i:=1 to nobj do ajoute(i,ones);
        for i:=1 to nobj do
                begin
                nullvec(faits[i]);
                ajoute(i,faits[i]);
                end;
        courant:=faits;
        for i:=1 to nbgroupes do
                begin
                temp1:=groupes[i];
                temp2:=temp1;
                difference(temp2,ones);
                for j:=1 to nobj do
                   if membre(j,temp1) then union(faits[j],temp2)
                   else                         union(faits[j],temp1);
                end;
        for i:=1 to nbgroupes do
                begin
                fini[i]:=cardinalite[i] = 0;
                pathlength[i]:=0;
                end;
        i:=0;
        repeat
                tousfinis:=true;
                i:=i+1;
                for j:=1 to nobj do
                        begin
                        temp1:=courant[j];
                        creefrontiere(courant[j],temp1);
                        difference(courant[j],faits[j]);
                        union(faits[j],temp1);
                        end;
                for j:=1 to nbgroupes do
                   if not(fini[j]) then
                        begin
                        fini[j]:=true;
                        for k:=1 to nobj do if membre(k,groupes[j]) then
                                begin
                                if not(vide(courant[k])) then
                                        begin
                                        fini[j]:=false;
                                        tousfinis:=false;
                                        pathlength[j]:=i;
                                        end;
                                end;
                        end;
        until tousfinis;
        end;
         
         
        FUNCTION maxdist(var mat:VecteurBit):REAL;
        var i,j:INTEGER;
            maxx,minx,maxy,miny:REAL;
         
                PROCEDURE minmax(var x,max,min:REAL);
                begin
                if x > max then max:=x
                else if x < min then min:=x;
                end;
         
                FUNCTION orthodromie(lat1,lon1,lat2,lon2:REAL):REAL;
                begin
                lat1:=lat1/rad;
                lat2:=lat2/rad;
                lon1:=(lon2-lon1)/rad;
                lat1:=(sin(lat1)*sin(lat2)+cos(lat1)*cos(lat2)*cos(lon1));
                if abs(lat1) > epsilon then
                        begin
                        lat2:=arctan(sqrt(1-lat1*lat1)/lat1);
                        if lat1 < -epsilon then lat2:=3.1415926535+lat2;
                        orthodromie:=lat2*60*rad;
                        end
                else orthodromie:=5400;
                end;  (* orthodromie *)
         
        begin (* MaxDist *)
        i:=0;
        repeat i:=i+1; until membre(i,mat);
        maxx:=position[1,i];minx:=maxx;
        maxy:=position[2,i];miny:=maxy;
        for j:=i+1 to nobj do if membre(j,mat) then
                begin
                minmax(position[1,j],maxx,minx);
                minmax(position[2,j],maxy,miny);
                end;
        maxdist:=orthodromie(maxx,maxy,minx,miny);
        end; (* MaxDist *)
         
        PROCEDURE ecritstatis;
        var i,j,compte:INTEGER;
            x:real;
         
                PROCEDURE count;
                begin
                compte:=compte+1;
                if (compte mod 7 = 0)and (compte <> nbgroupes) then
                        begin
                        writeln(statis);
                        write(statis,' ':10);
                        end;
                end;
         
        begin
        writeln(statis);
        if lang=1 then writeln(statis,' Carte no ',nbreussis:4)
        else writeln(statis,' Map no. ',nbreussis:4);
        writeln(statis);
        if 1 in statisset then
                begin
                if lang=1 then write(statis,'  DE      ')
                else write(statis,'  SD      ');
                compte:=0;
                for i:=1 to nbgroupes do
                        begin
                        if cardinalite[i]>1 then x:=maxdist(groupes[i])
                        else x:=0;
                        write(statis,x:10:3);
                        count;
                        if 4 in StatisSet then
                                begin
                                if NbReussis = 0 then
                                        begin
                                        OriginalDiam@[i]:=x;
                                        EQDiam@[i]:=1;
                                        PPDiam@[i]:=0;
                                        end
                                else
                                        begin
                                        if Abs(x-OriginalDiam@[i]) < Epsilon 
                                                then EQDiam@[i]:=EQDiam@[i]+1
                                        else if x < OriginalDiam@[i] 
                                                then PPDiam@[i]:=PPDiam@[i]+1;
                                        end;
                                end;
                        end;
                writeln(statis);
                end;
        if 2 in statisset then
                begin
                compte:=0;
                if lang=1 then write(statis,'  DNL     ')
                else write(statis,'  PL      ');
                longueur;
                for i:=1 to nbgroupes do
                        begin
                        j:=0;
                        if cardinalite[i] > 1 then j:=pathlength[i];
                        write(statis,j:10);
                        count;
                        if 5 in StatisSet then
                                begin
                                if NbReussis = 0 then
                                        begin
                                        OriginalPl@[i]:=j;
                                        EQPl@[i]:=1;
                                        PPPl@[i]:=0;
                                        end
                                else
                                        begin
                                        if j = OriginalPl@[i] 
                                                then EQPl@[i]:=EQPl@[i]+1
                                        else if j < OriginalPl@[i] 
                                                then PPPl@[i]:=PPPl@[i]+1;
                                        end;
                                end;
                        end;
                writeln(statis);
                end;
        if 3 in statisset then
                begin
                for i:=1 to nbvar do
                        begin
                        compte:=0;
                        if lang=1 then write(statis,'  SCE v',i:2,' ')
                        else write(statis,'  SSQ v',i:2,' ');
                        for j:=1 to nbgroupes do
                                begin
                                write(statis,' ',pgdist[i][j]:9:3);
                                count;
                                end;
                        writeln(statis,' ',sommessq[i]:9:3);
                        end;
                end;
        if 6 in statisset then
                begin
                if NbReussis = 0 then for i:=1 to NbGroupes do
                        for j:=1 to Nobj do Freq@[i][j]:=0
                else for i:=1 to NbGroupes do
                        for j:=1 to Nobj do if Membre(j,Groupes[i]) 
                            then Freq@[i][j]:=Freq@[i][j]+1;
                end;
        end;

        Procedure Vaudor(       VAR Copy1,Copy2,Seed:VecteurBit;
                                VAR CardGr,SeedCopy,Priorite:IntGr;
                                VAR PCoeff:pourGr;
                                VAR Deplacement:Neighbours;
                                VAR i:integer);
         
                PROCEDURE shellsort(var a:pourgr;var b:intgr;n: INTEGER);
                label 1;
                var i,j,k,l,m,o,x: INTEGER;
                    w:REAL;
                begin
                m:=(2*n-1) div 2;
                while m<>0 do
                        begin
                        k:=n-m;
                        for j:=1 to k do
                                begin
                                i:=j;
                                while i>=1 do
                                        begin
                                        if a[i+m]<=a[i] then goto 1;
                                        o:=i+m;
                                        w:=a[i];a[i]:=a[o];a[o]:=w;
                                        x:=b[i];b[i]:=b[o];b[o]:=x;
                                        i:=i-m;
                                        end;
                1:            end;
                        m:=m  div 2;
                        end;
                end;          (* fin de shellsort *)
                 
                PROCEDURE sked;
                var i:INTEGER;
                begin
                for i:=1 to nbgroupes do
                        begin
                        priorite[i]:=i;
                        pcoeff[i]:=cardgr[i]/(cardinalite[i]+1);
                        end;
                        shellsort(pcoeff,priorite,nbgroupes);
                end;
                 
                PROCEDURE essai;
                var   i,ii,j,compte,cardgri,nbpasses:INTEGER;
                        grcomplet,completbis:GrBit;
                        faitbis:VecteurBit;
                        x,y:REAL;
                        succes:boolean;
                 
                        procedure force;
                        Label 777;
                        var i,j,gr,AEnlever,ind,prof,groupe:integer;
                                termine:boolean;
                                temp,fait,faitbis:VecteurBit;
                                
                        begin
                        nbforce:=nbforce+1;
                        ind:=0;
                        repeat ind:=ind+1; 
                        until not(membre(ind,dejafait));(* trouve abs *)
                        groupe:=0;
                        repeat groupe:=groupe+1; until cardgr[groupe] > 0;
                        prof:=1;
                        FaitBis:=dejafait;
                        Ajoute(Ind,FaitBis);
                        deplacement[prof]:=groupes[groupe];
                        Fait:=groupes[groupe];
                        repeat
                                prof:=prof+1;
                                creefrontiere(deplacement[prof],
                                        deplacement[prof-1]);
                                Intersection(deplacement[prof],
                                        deplacement[prof],FaitBis);
                                difference(deplacement[prof],fait);
                                union(fait,deplacement[prof]);
                        until (membre(ind,Fait))or(Prof=Nobj);
                        if prof=Nobj then goto 777; (* Pas de chemin *)
                        gr:=Ind;
                        Nullvec(fait);
                        for i:=prof-1 downto 1 do
                                begin
                                intersection(temp,deplacement[i],voisins[gr]);
                                Difference(temp,fait);
                                Intersection(Temp,Temp,FaitBis);
                                premier(j,temp);
                                AEnlever:=adresse(j);
                                ajoute(gr,groupes[AEnlever]);
                                enleve(j,groupes[AEnlever]);
                                Ajoute(gr,fait);
                                gr:=j;
                                end;
                        Ajoute(j,groupes[AEnlever]);
                        ajoute(ind,dejafait);
                        cardgr[groupe]:=cardinalite[groupe]-
                                cardvect(groupes[groupe]);
                777:        end;
                 
                        PROCEDURE croissance
                        (var dejafait:VecteurBit;var grcomplet:GrBit;i:INTEGER);
                        var j:INTEGER;
                            temp:VecteurBit;
                        begin
                        cardgri:=cardgr[i];
                        copy1:=groupes[i];
                        creefrontiere(copy2,copy1);
                        difference(copy2,dejafait);
                        if not(vide(copy2)) then
                            begin
                            for j:=1 to nbgroupes do if j <> i then
                                begin
                                temp:=copy2;
                                union(temp,dejafait);
                                if inclusegal(groupes[j],temp) then
                                        difference(copy2,groupes[j]);
                                end;
                            end;
                        if not(vide(copy2)) then
                                begin
                                compte:=cardvect(copy2);
                                j:=0;
                                while compte > cardgri do
                                        begin
                                        j:=j+1;
                                        if membre(j,copy2) then
                                                begin
                                                compte:=compte-1;
                                                enleve(j,copy2);
                                                end;
                                        end;
                                if not(vide(copy2)) then
                                        begin
                                        action:=true;
                                        union(dejafait,copy2);
                                        union(groupes[i],copy2);
                                        if compte= cardgri 
                                                then grcomplet:=grcomplet+[i];
                                        cardgr[i]:=cardgri-compte;
                                        end;
                                end;
                        end;
                 
                        PROCEDURE corrige;
                        var i:INTEGER;
                        begin
                        completbis:=grcomplet;
                        nullvec(faitbis);
                        for i:=1 to nbgroupes do if not(i in grcomplet) then
                                union(faitbis,groupes[i]);
                        repeat
                            action:=false;
                            for i:=1 to nbgroupes do if not(i in grcomplet) 
                                    then croissance(faitbis,completbis,i);
                        until not(action);
                        for i:=1 to nbgroupes do if i in grcomplet then
                                        difference(groupes[i],faitbis);
                        grcomplet:=[];
                        for i:=1 to nbgroupes do
                                begin
                                cardgr[i]:=cardinalite[i]-cardvect(groupes[i]);
                                if cardgr[i]=0 then grcomplet:=grcomplet+[i];
                                end;
                        union(dejafait,faitbis);
                        end;
                 
                begin (* essai *)
                nullvec(dejafait);
                nullvec(seed);
                grcomplet:=[];
                for i:=1 to nbgroupes do
                        begin
                        priorite[i]:=i;
                        cardgr[i]:=cardinalite[i]-1;
                        pcoeff[i]:=cardinalite[i];
                        end;
                shellsort(pcoeff,priorite,nbgroupes);
                for i:=1 to nbgroupes do
                        begin
                        repeat
                                x:=ggubfs(dseed);
                                j:=trunc(x*nobj+1);
                        until not(membre(j,seed));
                        seedcopy[i]:=j;
                        ajoute(j,seed);
                        nullvec(groupes[i]);
                        ajoute(j,groupes[i]);
                        ajoute(j,dejafait);
                        if cardgr[i]=0 then grcomplet:=grcomplet+[i];
                        end;
                nbpasses:=-1;
                repeat
                nbpasses:=nbpasses+1;
                repeat
                        action:=false;
                        for ii:=1 to nbgroupes do
                                begin
                                i:=priorite[ii];
                                if not(i in grcomplet) 
                                        then croissance(dejafait,grcomplet,i);
                                end;
                        succes:=egal(dejafait,vectinit);
                        sked;
                until (succes)or(not(action));
                if (not(succes))and(nbpasses < maxpasses) then corrige;
                until (succes)or(nbpasses=maxpasses);
                if cardvect(dejafait)=nobj-NbForces then
                        begin
                        if connexe then
                                begin
                                force;
                                succes:=egal(dejafait,vectinit);
                                if succes 
                                        then nbforcereussis:=nbforcereussis+1;
                                end;
                        end;
                totalpasses:=totalpasses+nbpasses;
                if succes then
                        begin
                        if connexe then
                                begin
                                nbreussis:=nbreussis+1;
                                carte;
                                CalculeStatistiques;
                                if statisbool then ecritstatis;
                                end;
                        end
                else if coccinelle then carte;
                end;  (* essai*)
         
        begin (* Vaudor *)
        i:=0;
        repeat
                i:=i+1;
                essai;
                if i mod 20 = 0 then
                     if lang=1 then writeln(' Tentative no',i:4,' : nombre de ',
                     'cartes completees avec succes = ',nbreussis:3)
                     else writeln(' Attempt no.',i:4,
                     ' : number of successfully completed maps = ',nbreussis:3);
        until nbreussis = nbcartes;
        end; (* Vaudor *)


        Procedure Kim(  VAR        OriginalGroup      :IntObj;
                        VAR        MAP                :MapType);

                PROCEDURE SetUp;
        
                        PROCEDURE CreateNodes; 
                        (* Creates new nodes and initializes them *)
                        VAR        i,j: INTEGER;
                        BEGIN (* CreateNodes *)
                        FOR i:=1 TO Nobj DO
                                BEGIN
                                NEW(Map[i]);
                                Map[i]@.Degree := 0;
                                Map[i]@.Used        := FALSE;
                                Map[i]@.Name        := i;
                                FOR j:=1 TO MaxDegree DO Map[i]@.Link[j] := NIL;
                                END;
                        END; (* CreatNodes *)
        
                        PROCEDURE CreateMap; (* Sets up the original links *)
                        VAR        i,j,Bolt,Nut                : INTEGER;
                        BEGIN (* CreateMap *)
                        for i:=1 to Nobj do
                            for j:=1 to Nobj do if Membre(j,Voisins[i]) then
                                begin
                                Bolt := i;
                                Nut  := j;
                                Map[Bolt]@.Degree:= Map[Bolt]@.Degree + 1;
                                Map[Bolt]@.Link[Map[Bolt]@.Degree]:= Map[Nut];
                                Map[Nut]@.Degree:= Map[Nut]@.Degree + 1;
                                Map[Nut]@.Link[Map[Nut]@.Degree]:= Map[Bolt];
                                end;
                        END; (* CreateMap *)
        
                BEGIN (* SetUp *)
                        CreateNodes;
                        CreateMap;
                END; (* SetUp *)
        
                PROCEDURE Permutation;(* Drives the permutations generations *)
                TYPE
                        List        = RECORD
                                        Name        : INTEGER;
                                        Sect        : INTEGER;
                                        END;
                        TreePtr     = @Tree;
                        Tree        = RECORD
                                        Count       : INTEGER;
                                        Name        : INTEGER;
                                        Left        : TreePtr;
                                        Right       : TreePtr;
                                        Up          : TreePtr;
                                        END;
        
                VAR        Web,TempWeb,
                        Lost,Mother,
                        NullLink        : TreePtr;
                        OK              : BOOLEAN;
                        PartSet         : ARRAY[1..MaxGroupes] OF List;
                        Done,Aborted        : INTEGER;
        
                        PROCEDURE PartitionSetUp;
                        VAR i        : INTEGER;
                        BEGIN (* PartitionSetUp *)
                        FOR i:=1 TO NbGroupes DO
                                BEGIN
                                PartSet[i].Name := i;
                                PartSet[i].Sect := Cardinalite[i];
                                END;
                        END; (* PartitionSetUp *)
        
                        PROCEDURE Refresh; 
                        (* Sets all the used nodes of map to new *)
                        VAR i: INTEGER;
                        BEGIN (* Refresh *)
                        FOR i:=1 TO Nobj DO
                                Map[i]@.Used := FALSE;
                        END; (* Refresh *)
        
                        PROCEDURE SpinWeb; 
                        (* Finds a random spaning tree over the map *)
                        VAR  StartPoint,
                             LostLink,
                             NofLinked                : INTEGER;
        
                             FUNCTION Spider : INTEGER; 
                             (* Finds a random place to start *)
                             BEGIN (* Spider *)
                                Spider := TRUNC(Nobj*GGubfs(DSeed)) + 1;
                             END; (* Spider *)
        
                             PROCEDURE Locator
                             (VAR Child : INTEGER; VAR Node : TreePtr);
                             VAR StillLost                : BOOLEAN;
                                 LostChild                : INTEGER;
                                 Mom                      : TreePtr;
        
                                 PROCEDURE Hound(Target: TreePtr);
                                 VAR i,MapName                : INTEGER;
                                 BEGIN (* Hound *)
                                   IF Target@.Name >= 0 THEN
                                     BEGIN
                                     MapName := Map[Target@.Name]@.Name;
                                     FOR i:=1 TO Map[MapName]@.Degree DO
                                       BEGIN
                                       IF not Map[MapName]@.Link[i]@.Used THEN
                                         BEGIN
                                         LostChild := 
                                                 Map[MapName]@.Link[i]@.Name;
                                         StillLost := FALSE;
                                         Mom        := Target;
                                         END;
                                       END;
                                     END;
                                     IF StillLost AND (Target@.Left <> NIL) 
                                             THEN Hound(Target@.Left);
                                     IF StillLost AND (Target@.Right <> NIL) 
                                             THEN Hound(Target@.Right);
                                 END; (* Hound *)
        
                             BEGIN (* Locator *)
                             StillLost  := TRUE;
                             Hound(Web);
                             IF StillLost then
                             if lang=1 then WRITELN
                                (' Certaines localites n''ont aucun lien.',
                                ' Veuillez verifier votre fichier.')
                             else WRITELN(' Some localities do not have link',
                                ' edges. Please check your file.');
                             Child:= LostChild;
                             Node := Mom;
                             END; (* Locator *)
        
                             FUNCTION FindLink(Point : INTEGER): INTEGER;
                             (* Finds a random branch from a point of the map 
                             and returns to it *)
                             VAR i,Branch    : INTEGER;
                                 index,Nstick: INTEGER;
                                 Choice      : ARRAY[1..MaxDegree] OF INTEGER;
        
                             BEGIN (* FindLink *)
                             Nstick := Map[Point]@.Degree;
                             index  := 0;
                             FOR i:=1 TO Nstick DO
                                BEGIN
                                Branch := Map[Point]@.Link[i]@.Name;
                                IF not Map[Branch]@.Used THEN
                                        BEGIN
                                        index                := index + 1;
                                        Choice[index]        := Branch;
                                        END;
                                END;
                             IF index <> 0 THEN FindLink := 
                                     Choice[TRUNC(index*GGubfs(Dseed)) + 1]
                             ELSE FindLink := 0;
                             END; (* FindLink *)
        
                             PROCEDURE WebLink(Node : TreePtr; VAR N : INTEGER);
                             VAR NewLink        : INTEGER;
                             BEGIN (* WebLink *)
                             NewLink := FindLink(Node@.Name);
                             IF (NewLink <> 0) THEN
                                BEGIN
                                New(TempWeb);
                                          TempWeb@.Name   := NewLink;
                                        TempWeb@.Count    := 0;
                                        Node@.Left        := TempWeb;
                                        TempWeb@.Up       := Node;
                                        Map[NewLink]@.Used:= TRUE;
                                        N                 := N + 1;
                                        WebLink(TempWeb,N);
                                END
                             ELSE Node@.Left:= NIL;
                             NewLink := FindLink(Node@.Name);
                             IF NewLink <> 0 THEN
                                BEGIN
                                        New(TempWeb);
                                        TempWeb@.Name      := NewLink;
                                        TempWeb@.Count     := 0;
                                        Node@.Right        := TempWeb;
                                        TempWeb@.Up        := Node;
                                        Map[NewLink]@.Used := TRUE;
                                        N                  := N + 1;
                                        WebLink(TempWeb,N);
                                END
                             ELSE Node@.Right:= NIL;
                             END; (* WebLink *)
        
                        BEGIN (* SpinWeb *)
                        StartPoint          := Spider;
                        NEW(Web);
                        Web@.Name           := StartPoint;
                        Web@.Count          := 0;
                        Web@.Up             := NIL;
                        Map[Web@.Name]@.Used:= TRUE;
                        NofLinked           := 1;
                        WebLink(Web,NofLinked);
                        IF NofLinked < Nobj THEN (* There is a singularity *)
                          BEGIN
                          REPEAT
                                Locator(LostLink,Mother);
                                NEW(Lost);
                                Lost@.Name:= LostLink;
                                Lost@.Count:= 0;
                                Map[LostLink]@.Used := TRUE;
                                NofLinked:= NofLinked + 1;
                                WebLink(Lost,NofLinked); (* Now attach the 
                                lostlink and mother while creating a null *)
                                New(NullLink);
                                NullLink@.Name:= -1;
                                NullLink@.Count:= 0;
                                NullLink@.Up:= Mother;
                                Mother@.Left@.Up:= NullLink;
                                NullLink@.Left:= Mother@.Left;
                                Mother@.Left:= NullLink;
                                Lost@.Up:= NullLink;
                                NullLink@.Right:= Lost;
                          UNTIL (NofLinked >= Nobj);
                          END;
                        END; (* SpinWeb *)
        
                        PROCEDURE Enumerate(Node : TreePtr);
                        BEGIN (* Enumerate *)
                        IF (Node@.Left = NIL) AND (Node@.Right = NIL) 
                                THEN (* End *)
                                Node@.Count := 1
                        ELSE
                                BEGIN
                                IF Node@.Left <> NIL THEN
                                        Enumerate(Node@.Left);
                                IF Node@.Right <> NIL THEN
                                        Enumerate(Node@.Right);
                                END;
                        IF Node@.Name >= 0 THEN
                          BEGIN
                          IF (Node@.Left <> NIL) AND (Node@.Right <> NIL) THEN
                                    Node@.Count:= 
                                Node@.Left@.Count + Node@.Right@.Count + 1;
                          IF (Node@.Left <> NIL) AND (Node@.Right = NIL) THEN
                                Node@.Count := Node@.Left@.Count + 1;
                          IF (Node@.Left = NIL) AND (Node@.Right <> NIL) THEN
                                Node@.Count := Node@.Right@.Count + 1;
                          END
                        ELSE Node@.Count := 
                                Node@.Left@.Count + Node@.Right@.Count;
                        END; (* Enumerate *)
        
                        FUNCTION Assignment: BOOLEAN;
                        VAR Matched: BOOLEAN;
                            Neck: TreePtr;
                            NofMatched,
                            EndofSet,
                            Tag: INTEGER;
        
                            PROCEDURE Match
                            (VAR Hit: TreePtr; VAR Found : BOOLEAN;
                            VAR GroupName : INTEGER; Target : TreePtr);
                            VAR i : INTEGER;
                            BEGIN (* Match *)
                            IF Target@.Name >= 0 THEN
                                BEGIN
                                i := 1;
                                REPEAT
                                  IF Target@.Count = PartSet[i].Sect THEN
                                        BEGIN
                                        Found:= TRUE;
                                        Hit:= Target;
                                        GroupName:= PartSet[i].Name;
                                        PartSet[i].Name:= 
                                                PartSet[EndofSet].Name;
                                        PartSet[i].Sect:=
                                                PartSet[EndofSet].Sect;
                                        EndofSet:= EndofSet - 1;
                                        END
                                  ELSE  i := i + 1;
                                UNTIL ((Found) OR (i > EndofSet));
                                END;
                            IF (not Found) AND (Target@.Left <> NIL) THEN
                                Match(Hit,Found,GroupName,Target@.Left);
                            IF (not Found) AND (Target@.Right <> NIL) THEN
                                Match(Hit,Found,GroupName,Target@.Right);
                            END; (* Match *)
        
                            PROCEDURE Correct(Node : TreePtr);
                            VAR Correction: INTEGER;
                            BEGIN (* Correct *)
                            Correction := Node@.Count;
                            WHILE Node@.Up <> NIL DO
                                BEGIN
                                Node:= Node@.Up;
                                Node@.Count:= Node@.Count - Correction;
                                END;
                            END; (* Correct *)
        
                            PROCEDURE PickandName (Node:TreePtr;Josee:INTEGER);
                            BEGIN (* PickandName *)
                            IF Node@.Name >= 0 
                                    THEN OriginalGroup[Node@.Name] := Josee;
                            IF Node@.Left <> NIL 
                                    THEN PickandName(Node@.Left,Josee);
                            IF Node@.Right <> NIL 
                                    THEN PickandName(Node@.Right,Josee);
                            DISPOSE(Node);
                            END; (* PickandName *)
        
                        BEGIN (* Assignment *)
                        NofMatched := 0;
                        EndofSet   := NbGroupes;
                        New(Neck);
                        REPEAT
                                Matched := FALSE;
                                Match(Neck,Matched,Tag,Web);
                                IF Matched AND (Neck <> Web) THEN
                                  BEGIN
                                  Correct(Neck);
                                  IF Neck@.Up@.Left = Neck 
                                            THEN Neck@.Up@.Left  := NIL
                                  ELSE Neck@.Up@.Right := NIL;
                                  PickandName(Neck,Tag);
                                  END;
                        UNTIL ((not Matched) OR (Neck = Web));
                        IF Neck = Web THEN PickandName(Neck,Tag);
                        Assignment := Matched;
                        END; (* Assignment *)
        
                        PROCEDURE OutputGroups;
                        VAR i        : INTEGER;
                        BEGIN (* OutputGroups *)
                        for i:=1 to NbGroupes do Nullvec(Groupes[i]);
                        FOR i:=1 TO Nobj DO Ajoute(i,Groupes[OriginalGroup[i]]);
                        Carte;
                        CalculeStatistiques;
                        if statisbool then ecritstatis;
                        NbReussis:=NbReussis+1;
                        END; (* OutputGroups *)
        
                BEGIN (* Permutation *)
                Done    := 0;
                Aborted := 0;
                OutputGroups;        (* The original groups are outputed *)
                WHILE (Done < NbCartes) AND (Aborted <= MaxAbort) DO
                        BEGIN
                        PartitionSetUp;
                        Refresh;
                        SpinWeb;
                        Enumerate(Web);
                        OK := Assignment;
                        IF OK THEN
                                BEGIN
                                OutputGroups;
                                Done := Done + 1;
                                END
                        ELSE Aborted := Aborted + 1;
                        END;
                END; (* Permutation *)
      
        BEGIN (* Kim *)
        SetUp;
        Permutation;
        NbReussis:=NbReussis-1;
        end;  (* Kim *)

        Procedure GrandInit;
        var i,j,k,classe,compte:INTEGER;
                x,y,maxx,minx,maxy,miny,SommeX,SommeX2:REAL;

                PROCEDURE count2;
                begin
                compte:=compte+1;
                if (compte mod 7 = 0)and (compte <> nbgroupes) then
                        begin
                        writeln(statis);
                        write(statis,' ':10);
                        end;
                end;
                 
                PROCEDURE shell(var a:intobj;n: INTEGER);
                label 1;
                var i,j,k,l,m,o,w: INTEGER;
                begin
                m:=2*n-1 div 2;
                while m<>0 do
                        begin
                        k:=n-m;
                        for j:=1 to k do
                                begin
                                i:=j;
                                while i>=1 do
                                   begin
                                   if a[i+m]<=a[i] then goto 1;
                                   o:=i+m;
                                   w:=a[i];a[i]:=a[o];a[o]:=w;
                                   w:=coordx[1][i];
                                   coordx[1][i]:=coordx[1][o];coordx[1][o]:=w;
                                   w:=reference[i];
                                   reference[i]:=reference[o];reference[o]:=w;
                                   i:=i-m;
                                   end;
                1:            end;
                        m:=m  div 2;
                        end;
                end;          (* fin de shell *)
                 
                PROCEDURE lecture;
                var        i,j,k:INTEGER;
                        stop:boolean;
                 
                        PROCEDURE init(var k:INTEGER);
                        var i,j,l:INTEGER;
                        begin
                        j:=k div nbbits;
                        if j > nbmots then
                                begin
                                for l:= 1 to nobj do
                                     for i:=nbmots+1 to j do voisins[l][i]:=[];
                                nbmots:=j;
                                end;
                        for i:=nobj+1 to k do nullvec(voisins[i]);
                        nobj:=k;
                        end;
                 
                begin (* lecture *)
                reset(Links);
                if eof(Links) then erreur(2);
                NofLink := 0;
                if Not(VaudorMethod) then New(OriginalGroup);
                stop:=false;
                while not(stop) do
                        begin
                        if eof(Links) then stop:=true
                        else
                                begin
                                if Links@ in['0'..'9'] then
                                        begin
                                        read(Links,j,k);
                                        NofLink := NofLink + 1;
                                        if j > nobj then init(j);
                                        if k > nobj then init(k);
                                        ajoute(k,voisins[j]);
                                        ajoute(j,voisins[k]);
                                        end
                                else get(Links);
                                end;
                        end;
                nullvec(vectinit);
                for i:=1 to nobj do ajoute(i,vectinit);
                MaxPasses:=0;
                NbForces:=0;
                if VaudorMethod then
                   begin
                   if NbCartes > 0 then
                        begin
                        if lang=1 then begin
                        writeln(' Combien de fois peut-on tenter de deplacer',
                             ' les groupes ?');
                        writeln(' (Recommande: 3-5) ');
                        end else begin
                       writeln(' Number of times groups can be moved around ?');
                        writeln(' (Recommended: 3-5) ');
                        end;
                        readln(maxpasses);
                        if lang=1 then begin
                writeln(' Jusqu''a combien de points manquants doit-on tenter');
                writeln(' de completer une carte de force ?  (Recommande: 1)' );
                        end else begin
                writeln(' Until how many missing points do you want to force');
                writeln(' completing the maps ?  (Recommended: 1)');
                        end;
                        readln(NbForces);
                        end;
                   end
                else for i:=1 to Nobj do 
                        if CardVect(Voisins[i]) > Maxdegree then erreur(6);
                end;   (* lecture *)
        
        begin
        identification(output);
        totalpasses:=0;
        nbreussis:=0;
        nbforcereussis:=0;
        nbforce:=0;
        dseed    :=5.987345;
        nobj:=0;nbmots:=0;
        if lang=1 then
        writeln(' Combien de permutations de la carte faut-il realiser ?')
        else writeln(' How many permutations of the map do you want ? ');
        readln(nbcartes); 
        VaudorMethod:=true;
        if NbCartes > 0 then
           begin
           if lang=1 then begin
           writeln(' Initialisation du generateur de nombres aleatoires:');
           writeln(' Tapez un ENTIER entre 1 et 100. ');
           end else begin
           writeln(' Initialization of the random number generator: ');
           writeln(' Type an INTEGER between 1 and 100. ');
           end;
           readln(jj);
           for i:=1 to jj do x:=ggubfs(dseed);
           if lang=1 then begin 
           writeln
           (' Methode des cercles concentriques, plutot que l''arborescence',
           ' aleatoire ?');
           writeln(' (Tapez O pour la methode des cercles concentriques,'); 
           writeln('        N pour l''arborescence aleatoire.)');
           end else begin
           writeln(' RING method rather than RANDOM TREE ? (Y or N)');
           writeln(' (Type Y for Ring, or N for Random Tree.) ');
           end;
           VaudorMethod:=ValideBool;
           end;
        Lecture;
        GrapheSet:=[];
        if lang=1 then begin
        writeln(' FICHIER DES CARTES: ');
        writeln('    Dessiner les cartes ? (O ou N) ');
        end else begin
        writeln(' GRAPHICS FILE: ');
        writeln('    Print the maps ? (Y or N) ');
        end;
        if ValideBool then GrapheSet:=[1];
        if lang=1 then begin
        writeln('    Inscrire en representation codee quelles localites',
                ' forment chaque groupe ?');
        writeln('    (O ou N) ');
        end else
        writeln('    Print a coded representation of the localities',
                ' in each group ? (Y or N) ');
        if Validebool then GrapheSet:=GrapheSet+[2];
        if lang=1 then
        writeln('    Inscrire quel no de groupe a ete attribue a chaque',
        ' localite ? (O ou N) ')
        else writeln
        ('    Print the group attributed to every locality ? (Y or N) ');
        if ValideBool then GrapheSet:=GrapheSet+[3];
        Dessin:=GrapheSet<>[];
        coccinelle:=false;
        if dessin then
                begin
                rewrite(Graphics);
                identification(Graphics);
                if lang=1 then writeln(Graphics,
                    ' C O C O P A N -- Fichier des cartes')
                else writeln(Graphics,' C O C O P A N -- Map file');
                if ( VaudorMethod ) AND ( NbCartes > 0 ) then
                    begin
                    if lang=1 then begin
                    writeln('    DEBUG: Inscrire toutes les cartes et/ou les',
                    ' listes de bits,');
                    writeln('    meme pour les cartes rejetees ? (O ou N) ');
                    end else begin
                    writeln('    DEBUG: Print all maps and/or lists of bits,');
                    writeln('    even for the rejected maps ? (Y or N) ');
                    end;
                    coccinelle:=ValideBool;
                    end;
                if 1 in GrapheSet then
                    begin
                   if lang=1 then writeln
                    ('    Largeur des cartes (en nombre de caracteres) ? ')
                    else writeln
                    ('    Width of the maps (in number of characters) ? ');
                    readln(largeur);
                    end;
                end;
        if lang=1 then writeln(' Nombre de variables a analyser ? ')
        else writeln(' Number of variables to be analysed ? ');
        readln(nbvar);
        statisset:=[];
        if lang=1 then begin
        writeln(' FICHIER DES STATISTIQUES: ');
        writeln('    Diametre des ensembles (Set Diameter) ? (O ou N) ');
        end else begin
        writeln(' STATISTICS FILE: ');
        writeln('    Set Diameter (great circle distance) ? (Y or N) ');
        end;
        if validebool then
                begin
                statisset:=statisset+[1];
                if NbCartes > 0 then
                   begin
                   if lang=1 then writeln
                   ('    Probabilites des diametres d''ensembles ? (O ou N) ')
                   else writeln
                   ('    Probability of Set Diameter ? (Y or N) ');
                   if ValideBool then
                        begin
                        statisset:=statisset+[4];
                        New(OriginalDiam);
                        New(PPDiam);
                        New(EQDiam);
                        end;
                   end;
                end;
        if lang=1 then begin
        writeln('    Diametre en nombre de liens (Path Length) ? (O ou N) ');
        writeln('    (Attention: temps de calcul eleve s''il y a',
                ' beaucoup de points.)');
        end else begin
        writeln('    Path Length (graph diameter) ? (Y or N) ');
        writeln('    (Beware: high computing time for large data sets.)');
        end;
        if validebool then
                begin
                statisset:=statisset+[2];
                if NbCartes > 0 then
                        begin
                        if lang=1 then writeln('    Probabilites des',
                        ' diametres en nombre de liens ? (O ou N) ')
                        else writeln
                        ('    Probability of Path Length ? (Y or N) ');
                        if ValideBool then
                                begin
                                statisset:=statisset+[5];
                                New(OriginalPl);
                                New(PPPl);
                                New(EQPl);
                                end;
                        end;
                end;
        if lang=1 then begin
        writeln
        ('    Statistiques pour CHAQUE carte ? (O ou N) ');
        end else writeln('    Statistics for EACH map ? (Y or N) ');
        if validebool then statisset:=statisset+[3];
        if NbCartes > 0 then
                begin
                if lang=1 then writeln('    Statistiques quant a la frequence ',
                        'des localites dans chaque groupe? (O ou N)')
                else writeln('    Statistics about frequency ',
                        'of localities in each group ? (Y or N) ');
                if ValideBool then
                        begin
                        StatisSet:=StatisSet+[6];
                        New(Freq);
                        end;
                end;
        statisbool:=statisset <>[];
        reset(DataFile);
        if eof(DataFile) then erreur(1);
        nbgroupes:=0;
        for i:=1 to nobj do
                begin
                for j:=1 to nbvar do
                        begin
                        while (DataFile@ in[' '])
                                and(Not(eof(DataFile))) do get(DataFile);
                        if eof(DataFile) then erreur(4);
                        read(DataFile,donnee[j][i]);
                        end;
                while (DataFile@ in[' '])
                        and(Not(eof(DataFile))) do get(DataFile);
                if eof(DataFile) then erreur(4);
                read(DataFile,classe);
                if classe > nbgroupes then
                        begin
                        for j:=nbgroupes+1 to classe do
                                begin
                                cardinalite[j]:=0;
                                nullvec(groupes[j]);
                                end;
                        nbgroupes:=classe;
                        end;
                if Not(VaudorMethod) then OriginalGroup@[i]:=Classe;
                cardinalite[classe]:=cardinalite[classe]+1;
                ajoute(i,groupes[classe]);
                end;
        if statisbool  then
                begin
                rewrite(statis);
                Identification(Statis);
                if lang=1 then begin
                writeln(statis,' C O C O P A N - Fichier des statistiques');
                writeln(statis);
                writeln(statis,
                        ' Le fichier des liens contient: ');
                writeln(statis,'    Nombre total de liens:',NofLink:10);
                writeln(statis,'    Nombre moyen de liens:',NofLink/Nobj:10:5);
                end else begin
                writeln(statis,' C O C O P A N - Statistics file');
                writeln(statis);
                writeln(statis,
                        ' The file of link edges contains the following: ');
                writeln(statis,'    Number of edges      :',NofLink:10);
                writeln(statis,'    Mean number of edges :',NofLink/Nobj:10:5);
                end;
                SommeX:=0;SommeX2:=0;
                for i:=1 to Nobj do
                        begin
                        x:=cardvect(Voisins[i]);
                        SommeX:=SommeX+x;
                        SommeX2:=SommeX2+Sqr(x);
                        end;
                x:=Nobj;
                SommeX2:=(SommeX2-Sqr(SommeX)/x)/(x-1.0);
                if lang=1 then begin
                writeln(statis,'    Ecart type           :',Sqrt(SommeX2):10:5);
                writeln(statis,'    Variance du n. liens :',SommeX2:10:5);
                writeln(statis);
                writeln(statis,' DE  est le diametre de l''ensemble ',
                   '(pour chaque groupe, en min. d''arc)');
                writeln(statis,' DNL est le diametre en nombre de liens ',
                   '(pour chaque groupe)');
                writeln(statis,
                   ' SCE pour chaque groupe, puis pour',
                   ' l''ANOVA, pour les variables V(1) a V(n)');
                writeln(statis,
                   '    ');
                end else begin
                writeln(statis,'    Standard deviation   :',Sqrt(SommeX2):10:5);
                writeln(statis,'    Variance of edges    :',SommeX2:10:5);
                writeln(statis);
                writeln(statis,' SD  stands for Set Diameter ',
                        '(for each group, in min. of arc)');
                writeln(statis,' PL  stands for Path Length ',
                        '(for each group)');
                writeln(statis,
                        ' SSQ stands for Sum of SQuares (for each group, then ',
                        'for the whole ANOVA),');
                writeln(statis,'     for variables V(1) to V(n)');
                end;
                writeln(statis);
                write(statis,' N. loc.  ');
                compte:=0;
                for j:=1 to nbgroupes do
                        begin
                        write(statis,cardinalite[j]:10);
                        count2;
                        end;
                writeln(statis);
                end;
        if VaudorMethod then k:=1 else k:=0;
        for i:=1 to nbvar do
                begin
                sumssq[i]:=ssq(donnee[i]);
                refr[i]:=sumx2;
                for j:=1 to nbgroupes do begin ppm[i][j]:=0;eqm[i][j]:=k;end;
                pp[i]:=0;
                eq[i]:=k;
                end;
        pgdist:=refr;
        sommessq:=sumssq;
        if ((Dessin ) AND (1 in GrapheSet)) OR (1 in StatisSet) then
                begin
                reset(coord);
                if eof(coord) then erreur(3);
                maxx:=-maxreel;
                minx:=maxreel;
                miny:=maxreel;
                for i:=1 to nobj do
                        begin
                        while (coord@ in[' '])
                               and(Not(eof(coord))) do get(coord);
                        if eof(coord) then erreur(5);
                        read(coord,x);
                        while (coord@ in[' '])
                               and(Not(eof(coord))) do get(coord);
                        if eof(coord) then erreur(5);
                        read(coord,y);
                        if y > 180 then y:=y-360;
                        position[1,i]:=x;
                        position[2,i]:=y;
                        if x < minx then minx:=x
                        else if x > maxx then maxx:=x;
                        if y < miny then miny:=y;
                        end;
                coordx[1][0]:=0;coordx[2][0]:=0;
                coordx[1][Nobj+1]:=0;coordx[2][Nobj+1]:=0;
                if 1 in GrapheSet then
                        begin
                        maxx:=(maxx-minx)/(largeur-1);
                        maxy:=maxx*facteur;
                        for i:=1 to nobj do
                           begin
                           coordx[1][i]:=trunc((position[1,i]-minx)/maxx)+1;
                           coordx[2][i]:=trunc((position[2,i]-miny)/maxy)+1;
                           reference[i]:=i;
                           end;
                        shell(coordx[2],nobj);
                        end;
                end;
        end; (* GrandInit *)

begin  (* Prog princ *)
(* termin(input);                            For IBM mainframes only *)
GrandInit;
init;
nullvec(dejafait);
if VaudorMethod then
        begin
        Carte;
        if statisbool then ecritstatis;
        if NbCartes > 0 then
                begin
                New(Copy1);
                New(Copy2);
                New(Seed);
                New(CardGr);
                New(SeedCopy);
                New(Priorite);
                New(PCoeff);
                New(Deplacement);
                Vaudor(         Copy1@,Copy2@,Seed@,
                                CardGr@,SeedCopy@,Priorite@,
                                PCoeff@,
                                Deplacement@,
                                i);
                if lang=1 then 
                  begin
                  writeln(' N. succes / N. tentatives = ',nbreussis:4,'/',i:5);
                  writeln(' Nombre de fois ou il a fallu deplacer des',
                          ' groupes =',totalpasses:3);
                  writeln(' Nombre de cartes qui furent forcees = ',
                        nbforce:3);
                  writeln(' Parmi celles-ci, ',nbforcereussis:3,
                          ' ont ete completees avec succes');
                  end else begin
                  writeln(' N. of success / N. of attempts = ',nbreussis:4,
                          '/',i:5);
                  writeln(' Number of maps that had to be forced =',
                                totalpasses:3);
                  writeln(' Number of maps where a single point was missing = ',
                          nbforce:3);
                  writeln(' Of these, ',nbforcereussis:3,
                          ' were successfully completed');
                  end;
                end;
        end
else
     begin
     New(MAP);
     if lang=1 then 
       begin
       writeln
       (' Combien d''arbres aleatoires permettez-vous d''avorter avant que le');
       writeln
       (' programme ne s''arrete? On recommande 10*(N de permutations).');
       end else begin
       writeln
       (' How many random trees can be aborted before the program');
       writeln(' is stopped ?  Recommended: 10*(N. of permutations). ');
       end;
     Readln(MaxAbort);
     Kim(OriginalGroup@, MAP@);
     end;
if NbCartes > 0 then
        begin
        writeln;
        if lang=1 then writeln(' Probabilites des statistiques SCE: ')
        else writeln(' Probabilities for SSQ statistics: ');
        writeln;
        if lang=1 then
        writeln('  Variable  Groupe   Plus petits  Egaux  N.cartes Prob(H0)')
        else
        writeln('  Variable  Group     Smaller    Equal   N. maps  Prob(H0)');
        NbReussis:=NbReussis+1;
        if StatisBool then
                begin
                writeln(Statis);writeln(Statis);
                if lang=1 then
                writeln(Statis,' Probabilites des statistiques de SCE: ')
                else writeln(Statis,' Probabilities for SSQ statistics: ');
                writeln(Statis);
                if lang=1 then writeln(Statis,
                '  Variable  Groupe   Plus petits  Egaux  N.cartes Prob(H0)')
                else writeln(Statis,
                '  Variable  Group     Smaller    Equal   N. maps  Prob(H0)');
                end;
        for i:=1 to nbvar do
                begin
                for j:=1 to nbgroupes do
                        writeln(i:7,'         ',alfabet[j],ppm[i][j]:10,
                                eqm[i][j]:10,nbreussis:10,
                                (ppm[i][j]+eqm[i][j])/nbreussis:10:4);
                writeln(i:7,'     total',pp[i]:10,eq[i]:10,nbreussis:10,
                                (pp[i]+eq[i])/nbreussis:10:4);
                writeln;
                if StatisBool then
                        begin
                        for j:=1 to nbgroupes do
                        writeln(Statis,i:7,'         ',alfabet[j],ppm[i][j]:10,
                                eqm[i][j]:10,nbreussis:10,
                                (ppm[i][j]+eqm[i][j])/nbreussis:10:4);
                        writeln(Statis,i:7,'     total',pp[i]:10,eq[i]:10,
                                nbreussis:10,(pp[i]+eq[i])/nbreussis:10:4);
                        writeln(Statis);
                        end;
                end;
        if 4 In StatisSet then
                begin
                writeln(Statis);
                if lang=1 then
                writeln(Statis,' Probabilites des statistiques DE: ')
                else writeln(Statis,' Probabilities for SD statistics: ');
                writeln(Statis);
                if lang=1 then
                writeln(Statis,'      Groupe  Plus petits  Egaux Prob(H0)')
                else 
                writeln(Statis,'       Group    Smaller    Equal Prob(H0)');
                for i:=1 to NbGroupes do
                        writeln(Statis,i:10,PPDiam@[i]:10,EQDiam@[i]:10,
                                (PPDiam@[i]+EQDiam@[i])/NbReussis:10:4);
                writeln(Statis);
                end;
        if 5 In StatisSet then
                begin
                writeln(Statis);
                if lang=1 then
                writeln(Statis,' Probabilites des statistiques DNL: ')
                else writeln(Statis,' Probabilities for PL statistics: ');
                writeln(Statis);
                if lang=1 then
                writeln(Statis,'      Groupe  Plus petits  Egaux Prob(H0)')
                else 
                writeln(Statis,'       Group    Smaller    Equal Prob(H0)');
                for i:=1 to NbGroupes do
                        writeln(Statis,i:10,PPPl@[i]:10,EQPl@[i]:10,
                                (PPPl@[i]+EQPl@[i])/NbReussis:10:4);
                writeln(Statis);
                end;
        if 6 In StatisSet then
                begin
                for i:=1 to NbGroupes do
                    begin
                    if lang=1 then writeln
                    (Statis,' Frequence des localites dans le groupe ',i:1)
                    else 
                    writeln(Statis,' Frequency of data points in group ',i:1);
                    for j:=1 to Nobj do
                        begin
                        write(Statis,Freq@[i][j]:5);
                        if j mod 12 =0 then writeln(Statis);
                        end;
                    if Nobj Mod 12 <> 0 then writeln(Statis);
                    end;
                writeln(Statis);
                end;
        end;
if lang=1 then writeln(' Fin du programme.')
else writeln(' End of the program.');
end.
