PROGRAM LANCE(ENTREEB,SORTIEP,OUTPUT,INPUT);
LABEL 777;
CONST MAXNOBJ=300;       (* Maximum number of objects *)
      MAXLIM=44850;      (* MAXNOBJ*(MAXNOBJ-1) DIV 2 *)
      MAXDIM=512;
      MAXREEL=1.70E38;   (* largest real number on this computer *)
      NBMOTS=1;
      NBCAR=10;
      LANG=1;            (* Parametre de langue: 1 = FRANCAIS
                            Language parameter : 2 = ENGLISH  *)
(* ================================================================== *)
(*     ENTREEB is the input SIMIL file.                               *)
(*     SORTIEP contains informations (in binary form) as to the       *)
(*             clustering structure. This information is passed       *)
(*             to DENDRO that will draw the dendrogram.               *)
(*     INPUT and OUTPUT are the user's terminal.                      *)
(*
This new version of the program computes Ward's (1963) hierarchi-
cal clustering method, using the combinatorial algorithm as
described for instance in Jain & Dubes (1988), in Everitt
(1980) or in the SAS Statistics User's Guide (1985). In the
book of Everitt, the clustering level scale is an error sum
of squares statistic (i.e., a variance). Jain & Dubes (1988)
use the fusion squared distance as given by the combinatorial
algorithm; this measure is a linear transformation of that
in Everitt. In the SAS manuel, the clustering level scale,
called "semipartial R-squared", is the between-cluster sum of
squares divided by the total sum of squares; this is again a
linear transformation of Everitt's measure.
 
     Notice that all the above-mentioned measures are squared
distances. We use instead the SQUARE ROOT of the fusion
squared distance given by the combinatorial algorithm. On
the one hand, this produces a better-looking dendrogram; on
the other, this is the appropriate distance matrix to compare
to the original distance matrix in order to produce a correct
measure of cophenetic correlation.
                                     P. Legendre, April 1989.
References:
Everitt, B. 1980. Cluster analysis. Wiley, New York.
Jain, A.K. & R.C. Dubes. 1988. Algorithms for clustering data.
   Prentice-Hall, Englewood Cliffs, New Jersey.
SAS. 1985. SAS user's guide: statistics. SAS Institute Inc.,
   Cary, North Carolina.
Ward, J.H. Jr. 1963. Hierarchical grouping to optimize an
   objective function. J. Amer. Stat. Assoc. 58: 236-244.
                                                                      *)
(* ================================================================== *)
 
TYPE NOEUDTYPE=RECORD
     RESULTAT:REAL;
     PTRGAUCHE,PTRDROIT:INTEGER;
     END;
     MATPTR=PACKED ARRAY(.1..MAXNOBJ.)OF 0..1024;
     VARIABLE=RECORD CASE INTEGER OF
              1:(INT:INTEGER);
              2:(RE:REAL);
              3:(MATCAR:PACKED ARRAY[1..NBCAR] OF CHAR);
              END;
VAR I,J,K,COMPTEUR,MAX,MAX1,NOBJ,ANCIEN,ANCIENJ,POINTEURJ:INTEGER;
    X,PLUSPETIT,ALPHAJ,ALPHAM,BETA,GAMMA:REAL;
    NOEUD:NOEUDTYPE;
    AJAM,DISTANCE,POIDS,WARD:BOOLEAN;
    ANCIENPTR:MATPTR;
    MATCARD:ARRAY(.1..MAXNOBJ.) OF INTEGER;
    DEJAFAIT:PACKED ARRAY(.1..MAXNOBJ.)OF BOOLEAN;
    DONNEE:ARRAY(.1..MAXLIM.)OF REAL;
    ENTREEB:FILE OF VARIABLE;
    SORTIEP:FILE OF NOEUDTYPE;
 
PROCEDURE LECTURE(VAR NOBJ:INTEGER);
VAR I,J:INTEGER;
    TEM:REAL;
    TEMPORAIRE:VARIABLE;
BEGIN
RESET(ENTREEB);REWRITE(SORTIEP);
TEMPORAIRE:=ENTREEB@;
NOBJ:=TEMPORAIRE.INT;
COMPTEUR:=0;
IF NOBJ > MAXNOBJ THEN
      BEGIN IF LANG=1 THEN
      WRITELN(' TROP D OBJETS POUR LA CAPACITE DE CE PROGRAMME')
      ELSE WRITELN(' TOO MANY OBJECTS; RE-DIMENSION THE PROGRAM');
      GOTO 777;
      END;
GET(ENTREEB);
GET(ENTREEB);
FOR I:=1 TO 13 DO
   FOR J:=1 TO NBMOTS DO GET(ENTREEB);
TEMPORAIRE:=ENTREEB@;GET(ENTREEB);
IF TEMPORAIRE.INT<>0 THEN
   FOR I:=1 TO NOBJ DO
      FOR J:=1 TO NBMOTS DO GET(ENTREEB);
FOR I:=1 TO NOBJ DO
      BEGIN
      FOR J:=I+1 TO NOBJ DO
            BEGIN
            TEMPORAIRE:=ENTREEB@;GET(ENTREEB);
            IF DISTANCE THEN
            DONNEE(.(J-1)*(J-2) DIV 2+I.):=(1-TEMPORAIRE.RE)
            ELSE DONNEE(.(J-1)*(J-2) DIV 2+I.):=TEMPORAIRE.RE;
            END;
      END;
FOR I:=1 TO NOBJ DO MATCARD(.I.):=1;
END;
 
PROCEDURE CALCUL;
VAR I,J,K,L,WM,WJ:INTEGER;
   REFERENCE:REAL;
 
   FUNCTION ADRESSE(I,J:INTEGER):INTEGER;
   BEGIN
   ADRESSE:=(I-2)*(I-1) DIV 2+J;
   END;
 
   PROCEDURE CHERCHEMAX(VAR MAX,MAX1:INTEGER);
   VAR I,J,K:INTEGER;
       X:REAL;
   BEGIN
   X:=-(MAXREEL);
   FOR I:=2 TO NOBJ DO IF NOT DEJAFAIT(.I.) THEN
      BEGIN
      K:=(I-2)*(I-1) DIV 2;
      FOR J:=1 TO I-1 DO
         BEGIN
         K:=K+1;
         IF((DONNEE(.K.)>X)AND NOT (DEJAFAIT(.J.))) THEN
            BEGIN
            X:=DONNEE(.K.);
            MAX:=J;MAX1:=I;
            END;
         END;
      END;
   REFERENCE:=X;
   END;
 
   PROCEDURE FONCTION(K,L,KK:INTEGER);
   VAR WK,WKWMWJ:REAL;
   BEGIN
   IF WARD THEN                                           (* For Ward *)
      BEGIN
      WK:=MATCARD[KK];
      WKWMWJ:=WK+WM+WJ;
      ALPHAJ:=(WK+WJ)/WKWMWJ;
      ALPHAM:=(WK+WM)/WKWMWJ;
      BETA:=-WK/WKWMWJ;
   (* The square root of the fusion distance is transformed back to S *)
      DONNEE[K]:=1-SQRT(ALPHAJ*SQR(1-DONNEE[K])
                       +ALPHAM*SQR(1-DONNEE[L])
                       +BETA  *SQR(1-REFERENCE));
      END
   ELSE
   DONNEE[K]:=(1-ALPHAJ-ALPHAM-BETA)
      +ALPHAJ*DONNEE[K]+ALPHAM*DONNEE[L]+REFERENCE*BETA
      -GAMMA*ABS(DONNEE[L]-DONNEE[K]);
   END;
 
BEGIN   (* DEBUT PROCEDURE CALCUL *)
FOR I:=1 TO NOBJ DO DEJAFAIT(.I.):=FALSE;
FOR I:=1 TO NOBJ DO ANCIENPTR(.I.):=I;
COMPTEUR:=MAXDIM;
FOR I:=2 TO NOBJ DO
      BEGIN
      CHERCHEMAX(MAX,MAX1);
      WJ:=MATCARD(.MAX.);
      WM:=MATCARD(.MAX1.);
      IF POIDS THEN
            BEGIN
            ALPHAJ:=WJ/(WJ+WM);              (* For UPGMA and UPGMC *)
            ALPHAM:=WM/(WJ+WM);
            IF AJAM THEN BETA:=-ALPHAJ*ALPHAM;         (* For UPGMC *)
            END;
      FOR J:=MAX1+1 TO NOBJ DO
            IF NOT(DEJAFAIT[J]) THEN
               FONCTION(ADRESSE(J,MAX),ADRESSE(J,MAX1),J);
      L:=ADRESSE(MAX1,0);
      FOR J:=1 TO MAX-1 DO
            BEGIN
            L:=L+1;
            IF NOT(DEJAFAIT[J]) THEN FONCTION(ADRESSE(MAX,J),L,J);
            END;
      L:=L+1;
      FOR J:=MAX+1 TO MAX1-1 DO
            BEGIN
            L:=L+1;
            IF NOT(DEJAFAIT[J]) THEN FONCTION(ADRESSE(J,MAX),L,J);
            END;
      COMPTEUR:=COMPTEUR+1;
      WITH NOEUD DO
            BEGIN
            RESULTAT:=REFERENCE;
            IF ANCIENPTR(.MAX1.)<MAXNOBJ THEN
                  BEGIN
                  PTRGAUCHE:=ANCIENPTR(.MAX.);
                  PTRDROIT:=ANCIENPTR(.MAX1.);
                  END
            ELSE
                  BEGIN
                  PTRGAUCHE:=ANCIENPTR(.MAX1.);
                  PTRDROIT :=ANCIENPTR(.MAX .);
                  END;
            SORTIEP@:=NOEUD;PUT(SORTIEP);
            IF RESULTAT<PLUSPETIT THEN PLUSPETIT:=RESULTAT;
            END;
      ANCIENPTR(.MAX.):=COMPTEUR;
      ANCIENPTR(.MAX1.):=COMPTEUR;
      DEJAFAIT(.MAX1.):=TRUE;
      MATCARD(.MAX.):=WJ+WM;
      DONNEE(.ADRESSE(MAX1,MAX).):=-1;
      END;
NOEUD.RESULTAT:=PLUSPETIT;
SORTIEP@:=NOEUD;PUT(SORTIEP);
END;
 
BEGIN  (* PROG PRINC *)
WRITELN;
IF LANG=1 THEN BEGIN
WRITELN(' P R O G R A M M E   L A N C E ',
        '-- Modele general de groupement agglomeratif.');
WRITELN;
WRITELN(' Version 2.2b (Modifie pour SIMIL 3.0 / Inclut Ward)');
WRITELN(' Auteur: A. VAUDOR');
WRITELN;
WRITELN(' DESIREZ-VOUS');
WRITELN('      1- Association moyenne  (UPGMA)');
WRITELN('      2- Poids proportionnels (WPGMA)');
WRITELN('      3- Groupement centroide (UPGMC)');
WRITELN('      4- Groupement median    (WPGMC)');
WRITELN('      5- Methode de Ward (1963)      ');
WRITELN('      6- Autres groupements combinatoires');
END ELSE BEGIN
WRITELN(' P R O G R A M   L A N C E ',
                    '-- General agglomerative clustering model.');
WRITELN;
WRITELN(' Version 2.2b (Modified for SIMIL 3.0 / Includes Ward)');
WRITELN(' Author: A. VAUDOR');
WRITELN;
WRITELN(' DO YOU WANT ');
WRITELN('      1 - Unweighted arithmetic average clust. (UPGMA)');
WRITELN('      2 - Weighted arithmetic average clust.   (WPGMA)');
WRITELN('      3 - Unweighted centroid clustering       (UPGMC)');
WRITELN('      4 - Weighted centroid clustering         (WPGMC)');
WRITELN('      5 - Ward''s minimum variance clustering         ');
WRITELN('      6 - Other combinatorial clustering methods');
END;
READLN(I);
WARD:=FALSE;
POIDS:=FALSE;
AJAM:=FALSE;
BETA:=0; GAMMA:=0;
CASE I OF
   1: POIDS:=TRUE;
   2: BEGIN
      ALPHAJ:=0.5;
      ALPHAM:=0.5;
      END;
   3: BEGIN
      POIDS:=TRUE;
      AJAM:=TRUE;
      END;
   4: BEGIN
      ALPHAJ:=0.5;
      ALPHAM:=0.5;
      BETA:=-0.25;
      END;
   5: BEGIN
      WARD:=TRUE;
      END;
   6: BEGIN
      IF LANG=1
      THEN WRITELN(' ENTREZ  ALPHA[J], ALPHA[M], BETA, GAMMA ')
      ELSE WRITELN(' TYPE IN  ALPHA[J], ALPHA[M], BETA, GAMMA ');
      READLN(ALPHAJ,ALPHAM,BETA,GAMMA);
      END;
   END; (* End CASE *)
WRITELN;
DISTANCE:=FALSE;
IF LANG=1 THEN WRITELN
(' LA MATRICE D''ENTREE EST DE SIMILARITES OU DE DISTANCES? (S ou D)')
          ELSE WRITELN
(' IS THE INPUT FILE A SIMILARITY OR A DISTANCE FILE? (Write S or D)');
WHILE NOT (INPUT@ IN(.'S','s','D','d'.)) DO GET(INPUT);
DISTANCE:=INPUT@ IN (.'D','d'.); READLN;
IF LANG=1 THEN BEGIN
if distance then
   begin
   writeln(' Lecture de la matrice de DISTANCES.');
   writeln(' Les tests a posteriori ne sont tous corrects que ');
   writeln(' pour les matrices de SIMILARITE. Si vous les demandez,');
   writeln('  - la chaine des liens primaires sera donc erronee; ');
   writeln('  - le signe des correlations cophenetiques sera inverse;');
   writeln('  - la distance de Gower sera incorrecte; ');
   writeln('  - les mesures d''entropie seront correctes.');
   end
   else writeln(' Lecture de la matrice de similarites.');
END ELSE BEGIN
if distance then
   begin
   writeln(' Reading the DISTANCE matrix.');
   writeln(' The a posteriori tests are all correct only for ');
   writeln(' SIMILARITY matrices. If you request them nevertheless,');
   writeln('  - the minimum spanning tree will then be incorrect; ');
   writeln('  - cophenetic correlations will have the wrong sign; ');
   writeln('  - the Gower distance will be incorrect; ');
   writeln('  - entropy measures will be correct.');
   end
   else writeln(' Reading the similarity matrix.');
END;
WRITELN;
LECTURE(NOBJ);
REWRITE(SORTIEP);
PLUSPETIT:=MAXREEL;
CALCUL;
WRITELN;
777:IF LANG=1 THEN WRITELN(' Fin du groupement.')
              ELSE WRITELN(' End of the clustering.');
END.
